import os
import time
from .config import config, Configuration
from .utils import json_load, json_dump

# ---

def get_proton_path(version):
	return os.path.join(config.common_dir, f'Proton {version}')

def get_proton_exec_path(version):
	return os.path.join(get_proton_path(version), 'proton')

def get_prefix_path(prefix_name):
	return os.path.join(config.prefixes_dir, prefix_name)

def get_prefix_content_path(prefix_name):
	return os.path.join(get_prefix_path(prefix_name), 'pfx/drive_c')

def get_temporary_path():
	return os.path.join(config.temporary_dir, str(time.time()))

def get_prefix_config_path(prefix_name):
	return os.path.join(get_prefix_path(prefix_name), 'config.json')

def prefix_exists(prefix_name):
	return os.path.exists(get_prefix_path(prefix_name))

def proton_exists(version):
	return os.path.exists(get_proton_exec_path(version))

def config_load():
	return Configuration(json_load(config.config_path))

def config_save(content):
	json_dump(config.config_path, vars(content))
	
def prefix_config_load(prefix_name):
	return json_load(get_prefix_config_path(prefix_name))

def prefix_config_save(prefix_name, content):
	json_dump(get_prefix_config_path(prefix_name), content)


import json
import subprocess
import os
import re

# ---

# Source: https://stackoverflow.com/a/4836734

def natural_sort(l): 
    convert = lambda text: int(text) if text.isdigit() else text.lower() 
    alphanum_key = lambda key: [convert(c) for c in re.split('([0-9]+)', key)] 
    return sorted(l, key=alphanum_key)

# ---

def json_load(path):
	with open(path, 'r') as f:
		return json.load(f)

def json_dump(path, content):
	with open(path, 'w') as f:
		return json.dump(content, f)

# ---

def cmd_exec(command, env=None):
    if env is None:
    	env = {}
    env = {**env, **dict(os.environ)}
    subprocess.run(command, shell=True, env=env)

def exec_and_exit(*args, **kwargs):
	cmd_exec(*args, **kwargs)
	exit()

# ---

# Source: https://gist.github.com/garrettdreyfus/8153571#gistcomment-3519679

def yesno(question):
    prompt = f'{question} ? (y/n): '
    ans = input(prompt).strip().lower()
    if ans not in ['y', 'n']:
        print(f'{ans} is invalid, please try again...')
        return yesno(question)
    if ans == 'y':
        return True
    return False

# ---

def die(message):
	print(message)
	exit()


import os

# ---

class Configuration(object):
	def __init__(self, dictionary=None):
		if dictionary is not None:
			self.__dict__.update(dictionary)

# ---

HOME_DIR = os.path.expanduser("~")

config = Configuration()
config.common_dir = os.path.join(HOME_DIR, '.steam/debian-installation/steamapps/common')
config.protonenv_dir = os.path.join(HOME_DIR, '.protonenv')
config.prefixes_dir = os.path.join(config.protonenv_dir, 'prefixes')
config.temporary_dir = os.path.join(config.protonenv_dir, 'temp')
config.config_path = os.path.join(config.protonenv_dir, 'config.json')


import os
import shutil
import glob
import shlex

from .config import config
from .utils import *
from .core_utils import *
from .cli_utils import *

# ---

HOME_DIR = os.path.expanduser("~")
SCRIPT_DIR = os.path.dirname(os.path.realpath(__file__))
proton_appids = json_load(os.path.join(SCRIPT_DIR, 'proton.json'))

# ---

def proton_exec(version, prefix_name, command, flags=None):
	proton_exec_path = get_proton_exec_path(version)
	proton_exec_path = shlex.quote(proton_exec_path)
	command 		 = shlex.quote(command)
	if prefix_name is None:
		path = get_temporary_path()
		os.makedirs(path)
	else:
		path = get_prefix_path(prefix_name)
	env = {'STEAM_COMPAT_DATA_PATH': path}
	if flags is not None:
		for k,v in flags.items():
			env[k] = v
	cmd_exec(f'{proton_exec_path} waitforexitandrun {command}', env=env)

def ask_common_dir():
	while True:
		ans = input(f"Steam app common directory? (default: {config.common_dir}) [y to accept]\n").strip().lower()
		if ans == 'y':
			ans = config.common_dir
		if os.path.exists(ans):
			break
		print(f'{ans} does not exists, try again...')
	return ans


def setup():
	global config
	if not os.path.exists(config.protonenv_dir):
		print(f'Creating protonenv directory: {config.protonenv_dir}')
		os.makedirs(config.protonenv_dir)
	if not os.path.exists(config.prefixes_dir):
		print(f'Creating prefixes directory: {config.prefixes_dir}')
		os.makedirs(config.prefixes_dir)
	if not os.path.exists(config.temporary_dir):
		print(f'Creating temporary directory: {config.temporary_dir}')
		os.makedirs(config.temporary_dir)
	if not os.path.exists(config.config_path):
		config.common_dir = ask_common_dir()
		print(f'Creating config file: {config.config_path}')
		config_save(config)
	else:
		config = config_load()

def cleanup():
	for path in glob.glob(os.path.join(config.temporary_dir, '*')):
		assert HOME_DIR in path and '.protonenv' in path and config.protonenv_dir in path # be extra careful
		n_active = int(subprocess.check_output(f'lsof +d {path} | tail -n +2 | wc -l', shell=True))
		if n_active == 0:
			print(f'Cleaning up: {path}')
			shutil.rmtree(path)

def init():
	setup()
	cleanup()

# ---

@subcommand([])
def versions(args):
	for path in natural_sort(glob.glob(os.path.join(config.common_dir, 'Proton *'))):
		version = os.path.basename(path).split(' ')[-1]
		if not proton_exists(version):
			continue
		print(f'{version}')
	for path in natural_sort(glob.glob(os.path.join(config.prefixes_dir, '*'))):
		prefix_name = os.path.basename(path)
		prefix_config = prefix_config_load(prefix_name)
		print(f"{prefix_name} [{prefix_config['version']}]")

@subcommand([MutuallyExclusiveGroup([Argument("version", help="Proton version", nargs='?', default=None), 
									 Argument("--list", help="List installable Proton versions", action="store_true")], required=True)
			])
def install(args):
	if args.list:
		for version,_ in proton_appids.items():
			print(version, "\t[installed]" if proton_exists(version) else '')
		exit()
	if proton_exists(args.version):
		die(f'Proton {args.version} has already been installed.')
	appid = proton_appids.get(args.version)
	if appid is not None:
		exec_and_exit(f'steam steam://install/{appid}')
	die(f'Version {args.version} not found.')

@subcommand([Argument("name", help="Proton version or prefix name")])
def uninstall(args):
	appid = proton_appids.get(args.name)
	if appid is not None:
		exec_and_exit(f'steam steam://uninstall/{appid}')
	if not prefix_exists(args.name):
		die(f'Version {args.name} not found.')
	prefix_path = get_prefix_path(args.name)
	if yesno(f'Do you want to delete: {prefix_path}'):
		print(f'Deleting: {prefix_path}')
		shutil.rmtree(prefix_path)
	else:
		die('Nothing to do.')

@subcommand([Argument("version", help="Proton version"), Argument("prefix", help="Prefix name")])
def prefix(args):
	if prefix_exists(args.prefix):
		die(f'Prefix "{args.prefix}" has already exists.')
	if not proton_exists(args.version):
		die(f'Proton {args.version} does not exists.')
	os.makedirs(get_prefix_path(args.prefix))
	prefix_config = {"version": args.version}
	prefix_config_save(args.prefix, prefix_config)
	proton_exec(args.version, args.prefix, '') # initialize Proton prefix

@subcommand([Argument("prefix", help="Prefix name"), Argument("--flags", help="Environment variable: KEY=VALUE", nargs='+'), Argument("--command", help="Default command")])
def default(args):
	if not prefix_exists(args.prefix):
		die(f'Prefix {args.prefix} does not exists.')
	prefix_config = prefix_config_load(args.prefix)
	if args.command:
		prefix_config['command'] = args.command
		print(f'\"{prefix_config["command"]}\" will be run by default')
	if args.flags:
		prefix_config['flags'] = {}
		for flag in args.flags:
			try:
				k, v = flag.split('=')
			except:
				die('Invalid flag, it must be in the form of KEY=VALUE.')
			prefix_config['flags'][k] = v
		print(f'{prefix_config["flags"]} is set by default')
	if args.command or args.flags:
		prefix_config_save(args.prefix, prefix_config)
	else:
		die('Nothing to do.')

@subcommand([Argument("prefix", help="Prefix name")])
def info(args):
	if not prefix_exists(args.prefix):
		die(f'Prefix {args.prefix} does not exists.')
	prefix_config = prefix_config_load(args.prefix)
	die(f'\
{args.prefix}\n\
{"-"*len(args.prefix)}\n\
Proton\t: {prefix_config["version"]}\n\
Command\t: {prefix_config.get("command")}\n\
Flags\t: {prefix_config.get("flags")}\
		')

@subcommand([Argument("prefix", help="Prefix name"), Argument("--open", action="store_true", help="Open the directory")])
def directory(args):
	if not prefix_exists(args.prefix):
		die(f'Prefix {args.prefix} does not exists.')
	content_path = get_prefix_content_path(args.prefix)
	if args.open:
		cmd_exec(f'xdg-open {content_path}')
	die(content_path)

@subcommand([Argument("name", help="Proton version or prefix name"), Argument("command", nargs='?', default=None, help="Command to execute")])
def run(args):
	if not prefix_exists(args.name):
		if not proton_exists(args.name):
			die(f'{args.name} does not exists.')
		if not args.command:
			die('Command is not specified.')
		proton_exec(args.name, None, args.command)
		exit()
	prefix_config = prefix_config_load(args.name)
	command = args.command or prefix_config.get('command')
	if command is None:
		die("Command is not specified and the default command is also not set.")
	flags = prefix_config.get('flags')
	version = prefix_config.get('version')
	proton_exec(version, args.name, command, flags)

def main():
	init()
	args = cli.parse_args()
	if args.subcommand is None:
		cli.print_help()
	else:
		args.func(args)


import os
import argparse
import sys

# ---

# Reference: https://gist.github.com/mivade/384c2c41c3a29c637cb6c603d4197f9f

cli = argparse.ArgumentParser(description='Simple Proton version and prefix management.')
subparsers = cli.add_subparsers(dest="subcommand")

class Argument:
	def __init__(self, *name_or_flags, **kwargs):
		self.name_or_flags = list(name_or_flags)
		self.kwargs = kwargs

class MutuallyExclusiveGroup(Argument):
	def __init__(self, arguments, *name_or_flags, **kwargs):
		super().__init__(*name_or_flags, **kwargs)
		self.arguments = arguments

def subcommand(args=[], parent=subparsers, name=None):
    def decorator(func):
        parser = parent.add_parser(name or func.__name__, description=func.__doc__)
        for arg in args:
        	if isinstance(arg, MutuallyExclusiveGroup):
        		group = parser.add_mutually_exclusive_group(*arg.name_or_flags, **arg.kwargs)
        		for _arg in arg.arguments:
        			group.add_argument(*_arg.name_or_flags, **_arg.kwargs)
        	else:
        		parser.add_argument(*arg.name_or_flags, **arg.kwargs)
        parser.set_defaults(func=func)
    return decorator


import io
import os
import sys
from shutil import rmtree

from setuptools import find_packages, setup, Command

# ---

# SOURCE: https://github.com/navdeep-G/setup.py

# ---

# METADATA
NAME = 'protonenv'
DESCRIPTION = 'Simple Proton version and prefix management.'
URL = 'https://github.com/rizkiarm/protonenv'
EMAIL = 'rizki@rizkiarm.com'
AUTHOR = 'Muhammad Rizki Aulia Rahman Maulana'
REQUIRES_PYTHON = '>=3.6.0'
KEYWORDS = ['steam', 'proton', 'prefix', 'wine']

# ---

# README
here = os.path.abspath(os.path.dirname(__file__))
with io.open(os.path.join(here, 'README.md'), encoding='utf-8') as f:
    README = f.read()

# VERSION
about = {}
project_slug = NAME.lower().replace("-", "_").replace(" ", "_")
with open(os.path.join(here, project_slug, '__version__.py')) as f:
    exec(f.read(), about)
VERSION = about['__version__']

# ---

class UploadCommand(Command):
    """Support setup.py upload."""

    description = 'Build and publish the package.'
    user_options = []

    @staticmethod
    def status(s):
        """Prints things in bold."""
        print('\033[1m{0}\033[0m'.format(s))

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        try:
            self.status('Removing previous builds…')
            rmtree(os.path.join(here, 'dist'))
        except OSError:
            pass

        self.status('Building Source and Wheel (universal) distribution…')
        os.system('{0} setup.py sdist bdist_wheel --universal'.format(sys.executable))
        
        ans = input("Are you sure you want to publish?\n").strip().lower()
        if ans not in ['y']: exit()

        self.status('Uploading the package to PyPI via Twine…')
        os.system('twine upload dist/*')

        self.status('Pushing git tags…')
        os.system('git tag v{0}'.format(about['__version__']))
        os.system('git push --tags')

        sys.exit()

# ---

setup(
    name = NAME,
    version = VERSION,
    description = DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    author = AUTHOR,
    author_email = EMAIL,
    url = URL,
    keywords = KEYWORDS,
    license="MIT",
    classifiers = [
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
    ],
    packages = find_packages(exclude=["tests", "*.tests", "*.tests.*", "tests.*"]),
    include_package_data=True,
    entry_points={
        "console_scripts": [
            "protonenv=protonenv:main",
        ],
    },
    python_requires=REQUIRES_PYTHON,
    # $ setup.py publish support.
    cmdclass={
        'upload': UploadCommand,
    },
)